@@ -2,7 +2,7 @@
  * paths.c:   a path manipulation library using svn_stringbuf_t
  *
  * ====================================================================
- * Copyright (c) 2000-2004 CollabNet.  All rights reserved.
+ * Copyright (c) 2000-2007 CollabNet.  All rights reserved.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution.  The terms
@@ -76,6 +76,10 @@ svn_path_local_style(const char *path, apr_pool_t *pool)
   if (SVN_PATH_IS_EMPTY(path))
     return ".";
 
+  /* If PATH is a URL, the "local style" is the same as the input. */
+  if (svn_path_is_url(path))
+    return apr_pstrdup(pool, path);
+
   if ('/' != SVN_PATH_LOCAL_SEPARATOR)
     {
       char *p = apr_pstrdup(pool, path);
@@ -98,7 +102,8 @@ is_canonical(const char *path,
              apr_size_t len)
 {
   return (! SVN_PATH_IS_PLATFORM_EMPTY(path, len)
-          && (len <= 1 || path[len-1] != '/'));
+          && (svn_dirent_is_root(path, len) ||
+              (len <= 1 || path[len-1] != '/')));
 }
 #endif
 
@@ -281,7 +286,7 @@ svn_path_component_count(const char *path)
         ++path;
 
       start = path;
-      
+
       while (*path && *path != '/')
         ++path;
 
@@ -292,6 +297,7 @@ svn_path_component_count(const char *path)
   return count;
 }
 
+
 /* Return the length of substring necessary to encompass the entire
  * previous path segment in PATH, which should be a LEN byte string.
  *
@@ -315,8 +321,9 @@ previous_segment(const char *path,
     return len;
 }
 
+
 void
-svn_path_add_component(svn_stringbuf_t *path, 
+svn_path_add_component(svn_stringbuf_t *path,
                        const char *component)
 {
   apr_size_t len = strlen(component);
@@ -346,6 +353,7 @@ svn_path_remove_component(svn_stringbuf_t *path)
   path->data[path->len] = '\0';
 }
 
+
 void
 svn_path_remove_components(svn_stringbuf_t *path, apr_size_t n)
 {
@@ -389,7 +397,6 @@ svn_path_basename(const char *path, apr_pool_t *pool)
 }
 
 
-
 void
 svn_path_split(const char *path,
                const char **dirpath,
@@ -418,6 +425,49 @@ svn_path_is_empty(const char *path)
 }
 
 
+/* We decided against using apr_filepath_root here because of the negative
+   performance impact (creating a pool and converting strings ). */
+svn_boolean_t
+svn_dirent_is_root(const char *dirent, apr_size_t len)
+{
+  /* directory is root if it's equal to '/' */
+  if (len == 1 && dirent[0] == '/')
+    return TRUE;
+
+#if defined(WIN32) || defined(__CYGWIN__)
+  /* On Windows and Cygwin, 'H:' or 'H:/' (where 'H' is any letter)
+     are also root directories */
+  if ((len == 2 || len == 3) &&
+      (dirent[1] == ':') &&
+      ((dirent[0] >= 'A' && dirent[0] <= 'Z') ||
+       (dirent[0] >= 'a' && dirent[0] <= 'z')) &&
+      (len == 2 || (dirent[2] == '/' && len == 3)))
+    return TRUE;
+
+  /* On Windows and Cygwin, both //drive and //drive//share are root
+     directories */
+  if (len >= 2 && dirent[0] == '/' && dirent[1] == '/'
+      && dirent[len - 1] != '/')
+    {
+      int segments = 0;
+      int i;
+      for (i = len; i >= 2; i--)
+        {
+          if (dirent[i] == '/')
+            {
+              segments ++;
+              if (segments > 1)
+                return FALSE;
+            }
+        }
+      return (segments <= 1);
+    }
+#endif /* WIN32 or Cygwin */
+
+  return FALSE;
+}
+
+
 int
 svn_path_compare_paths(const char *path1,
                        const char *path2)
@@ -436,7 +486,7 @@ svn_path_compare_paths(const char *path1,
 
   /* Are the paths exactly the same? */
   if ((path1_len == path2_len) && (i >= min_len))
-    return 0;    
+    return 0;
 
   /* Children of paths are greater than their parents, but less than
      greater siblings of their parents. */
@@ -450,18 +500,20 @@ svn_path_compare_paths(const char *path1,
     return 1;
 
   /* Common prefix was skipped above, next character is compared to
-     determine order */
-  return path1[i] < path2[i] ? -1 : 1;
+     determine order.  We need to use an unsigned comparison, though,
+     so a "next character" of NULL (0x00) sorts numerically
+     smallest. */
+  return (unsigned char)(path1[i]) < (unsigned char)(path2[i]) ? -1 : 1;
 }
 
 
-/* Return the string length of the longest common ancestor of PATH1 and PATH2.  
+/* Return the string length of the longest common ancestor of PATH1 and PATH2.
  *
- * This function handles everything except the URL-handling logic 
- * of svn_path_get_longest_ancestor, and assumes that PATH1 and 
- * PATH2 are *not* URLs.  
+ * This function handles everything except the URL-handling logic
+ * of svn_path_get_longest_ancestor, and assumes that PATH1 and
+ * PATH2 are *not* URLs.
  *
- * If the two paths do not share a common ancestor, return 0. 
+ * If the two paths do not share a common ancestor, return 0.
  *
  * New strings are allocated in POOL.
  */
@@ -473,7 +525,7 @@ get_path_ancestor_length(const char *path1,
   apr_size_t path1_len, path2_len;
   apr_size_t i = 0;
   apr_size_t last_dirsep = 0;
-  
+
   path1_len = strlen(path1);
   path2_len = strlen(path2);
 
@@ -493,6 +545,12 @@ get_path_ancestor_length(const char *path1,
         break;
     }
 
+  /* two special cases:
+     1. '/' is the longest common ancestor of '/' and '/foo'
+     2. '/' is the longest common ancestor of '/rif' and '/raf' */
+  if (i == 1 && path1[0] == '/' && path2[0] == '/')
+    return 1;
+
   /* last_dirsep is now the offset of the last directory separator we
      crossed before reaching a non-matching byte.  i is the offset of
      that non-matching byte. */
@@ -501,6 +559,8 @@ get_path_ancestor_length(const char *path1,
            || ((i == path1_len) && (i == path2_len)))
     return i;
   else
+    if (last_dirsep == 0 && path1[0] == '/' && path2[0] == '/')
+      return 1;
     return last_dirsep;
 }
 
@@ -514,9 +574,9 @@ svn_path_get_longest_ancestor(const char *path1,
   path1_is_url = svn_path_is_url(path1);
   path2_is_url = svn_path_is_url(path2);
 
-  if (path1_is_url && path2_is_url) 
+  if (path1_is_url && path2_is_url)
     {
-      apr_size_t path_ancestor_len; 
+      apr_size_t path_ancestor_len;
       apr_size_t i = 0;
 
       /* Find ':' */
@@ -524,10 +584,10 @@ svn_path_get_longest_ancestor(const char *path1,
         {
           /* No shared protocol => no common prefix */
           if (path1[i] != path2[i])
-            return apr_pmemdup(pool, SVN_EMPTY_PATH, 
+            return apr_pmemdup(pool, SVN_EMPTY_PATH,
                                sizeof(SVN_EMPTY_PATH));
 
-          if (path1[i] == ':') 
+          if (path1[i] == ':')
             break;
 
           /* They're both URLs, so EOS can't come before ':' */
@@ -538,18 +598,19 @@ svn_path_get_longest_ancestor(const char *path1,
 
       i += 3;  /* Advance past '://' */
 
-      path_ancestor_len = get_path_ancestor_length(path1 + i, path2 + i, 
+      path_ancestor_len = get_path_ancestor_length(path1 + i, path2 + i,
                                                    pool);
 
-      if (path_ancestor_len == 0)
+      if (path_ancestor_len == 0 ||
+          (path_ancestor_len == 1 && (path1 + i)[0] == '/'))
         return apr_pmemdup(pool, SVN_EMPTY_PATH, sizeof(SVN_EMPTY_PATH));
       else
-        return apr_pstrndup(pool, path1, path_ancestor_len + i); 
+        return apr_pstrndup(pool, path1, path_ancestor_len + i);
     }
 
   else if ((! path1_is_url) && (! path2_is_url))
-    { 
-      return apr_pstrndup(pool, path1, 
+    {
+      return apr_pstrndup(pool, path1,
                           get_path_ancestor_length(path1, path2, pool));
     }
 
@@ -578,7 +639,8 @@ svn_path_is_child(const char *path1,
           || path2[0] == '/')                  /* "/foo" not a child */
         return NULL;
       else
-        return apr_pstrdup(pool, path2);      /* everything else is child */
+        /* everything else is child */
+        return pool ? apr_pstrdup(pool, path2) : path2;
     }
 
   /* Reach the end of at least one of the paths.  How should we handle
@@ -599,15 +661,16 @@ svn_path_is_child(const char *path1,
   if (path1[i] == '\0' && path2[i])
     {
       if (path2[i] == '/')
-        return apr_pstrdup(pool, path2 + i + 1);
+        return pool ? apr_pstrdup(pool, path2 + i + 1) : path2 + i + 1;
       else if (i == 1 && path1[0] == '/')
-        return apr_pstrdup(pool, path2 + 1);
+        return pool ? apr_pstrdup(pool, path2 + 1) : path2 + 1;
     }
 
   /* Otherwise, path2 isn't a child. */
   return NULL;
 }
 
+
 svn_boolean_t
 svn_path_is_ancestor(const char *path1, const char *path2)
 {
@@ -629,13 +692,15 @@ svn_path_is_ancestor(const char *path1, const char *path2)
 
   return FALSE;
 }
+
+
 apr_array_header_t *
 svn_path_decompose(const char *path,
                    apr_pool_t *pool)
 {
   apr_size_t i, oldi;
 
-  apr_array_header_t *components = 
+  apr_array_header_t *components =
     apr_array_make(pool, 1, sizeof(const char *));
 
   /* assert (is_canonical (path, strlen (path)));  ### Expensive strlen */
@@ -649,7 +714,7 @@ svn_path_decompose(const char *path,
     {
       char dirsep = '/';
 
-      *((const char **) apr_array_push(components))
+      APR_ARRAY_PUSH(components, const char *)
         = apr_pstrmemdup(pool, &dirsep, sizeof(dirsep));
 
       i++;
@@ -663,9 +728,9 @@ svn_path_decompose(const char *path,
       if ((path[i] == '/') || (path[i] == '\0'))
         {
           if (SVN_PATH_IS_PLATFORM_EMPTY(path + oldi, i - oldi))
-            *((const char **) apr_array_push(components)) = SVN_EMPTY_PATH;
+            APR_ARRAY_PUSH(components, const char *) = SVN_EMPTY_PATH;
           else
-            *((const char **) apr_array_push(components))
+            APR_ARRAY_PUSH(components, const char *)
               = apr_pstrmemdup(pool, path + oldi, i - oldi);
 
           i++;
@@ -680,6 +745,53 @@ svn_path_decompose(const char *path,
 }
 
 
+const char *
+svn_path_compose(const apr_array_header_t *components,
+                 apr_pool_t *pool)
+{
+  apr_size_t *lengths = apr_palloc(pool, components->nelts*sizeof(*lengths));
+  apr_size_t max_length = components->nelts;
+  char *path;
+  char *p;
+  int i;
+
+  /* Get the length of each component so a total length can be
+     calculated. */
+  for (i = 0; i < components->nelts; ++i)
+    {
+      apr_size_t l = strlen(APR_ARRAY_IDX(components, i, const char *));
+      lengths[i] = l;
+      max_length += l;
+    }
+
+  path = apr_palloc(pool, max_length + 1);
+  p = path;
+
+  for (i = 0; i < components->nelts; ++i)
+    {
+      /* Append a '/' to the path.  Handle the case with an absolute
+         path where a '/' appears in the first component.  Only append
+         a '/' if the component is the second component that does not
+         follow a "/" first component; or it is the third or later
+         component. */
+      if (i > 1 ||
+          (i == 1 && strcmp("/", APR_ARRAY_IDX(components,
+                                               0,
+                                               const char *)) != 0))
+        {
+          *p++ = '/';
+        }
+
+      memcpy(p, APR_ARRAY_IDX(components, i, const char *), lengths[i]);
+      p += lengths[i];
+    }
+
+  *p = '\0';
+
+  return path;
+}
+
+
 svn_boolean_t
 svn_path_is_single_path_component(const char *name)
 {
@@ -703,13 +815,13 @@ svn_boolean_t
 svn_path_is_backpath_present(const char *path)
 {
   int len = strlen(path);
-  
+
   if (! strcmp(path, ".."))
     return TRUE;
 
   if (! strncmp(path, "../", 3))
     return TRUE;
-  
+
   if (strstr(path, "/../") != NULL)
     return TRUE;
 
@@ -732,9 +844,10 @@ skip_uri_scheme(const char *path)
 {
   apr_size_t j;
 
-  for (j = 0; path[j]; ++j)
-    if (path[j] == ':' || path[j] == '/')
-       break;
+  /* A scheme is terminated by a : and cannot contain any /'s. */
+  for (j = 0; path[j] && path[j] != ':'; ++j)
+    if (path[j] == '/')
+      return NULL;
 
   if (j > 0 && path[j] == ':' && path[j+1] == '/' && path[j+2] == '/')
     return path + j + 3;
@@ -742,12 +855,13 @@ skip_uri_scheme(const char *path)
   return NULL;
 }
 
-svn_boolean_t 
+
+svn_boolean_t
 svn_path_is_url(const char *path)
 {
   /* ### This function is reaaaaaaaaaaaaaally stupid right now.
      We're just going to look for:
- 
+
         (scheme)://(optional_stuff)
 
      Where (scheme) has no ':' or '/' characters.
@@ -762,7 +876,7 @@ svn_path_is_url(const char *path)
 /* Here is the BNF for path components in a URI. "pchar" is a
    character in a path component.
 
-      pchar       = unreserved | escaped | 
+      pchar       = unreserved | escaped |
                     ":" | "@" | "&" | "=" | "+" | "$" | ","
       unreserved  = alphanum | mark
       mark        = "-" | "_" | "." | "!" | "~" | "*" | "'" | "(" | ")"
@@ -770,7 +884,7 @@ svn_path_is_url(const char *path)
    Note that "escaped" doesn't really apply to what users can put in
    their paths, so that really means the set of characters is:
 
-      alphanum | mark | ":" | "@" | "&" | "=" | "+" | "$" | "," 
+      alphanum | mark | ":" | "@" | "&" | "=" | "+" | "$" | ","
 */
 static const char uri_char_validity[256] = {
   0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,
@@ -797,7 +911,8 @@ static const char uri_char_validity[256] = {
   0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,
 };
 
-svn_boolean_t 
+
+svn_boolean_t
 svn_path_is_uri_safe(const char *path)
 {
   apr_size_t i;
@@ -834,11 +949,12 @@ svn_path_is_uri_safe(const char *path)
         {
           return FALSE;
         }
-    } 
+    }
 
   return TRUE;
 }
-  
+
+
 /* URI-encode each character c in PATH for which TABLE[c] is 0.
    If no encoding was needed, return PATH, else return a new string allocated
    in POOL. */
@@ -863,9 +979,9 @@ uri_escape(const char *path, const char table[], apr_pool_t *pool)
       /* First things first, copy all the good stuff that we haven't
          yet copied into our output buffer. */
       if (i - copied)
-        svn_stringbuf_appendbytes(retstr, path + copied, 
+        svn_stringbuf_appendbytes(retstr, path + copied,
                                   i - copied);
-      
+
       /* Now, sprintf() in our escaped character, making sure our
          buffer is big enough to hold the '%' and two digits.  We cast
          the C to unsigned char here because the 'X' format character
@@ -894,6 +1010,7 @@ uri_escape(const char *path, const char table[], apr_pool_t *pool)
   return retstr->data;
 }
 
+
 const char *
 svn_path_uri_encode(const char *path, apr_pool_t *pool)
 {
@@ -935,7 +1052,7 @@ svn_path_uri_from_iri(const char *iri, apr_pool_t *pool)
   return uri_escape(iri, iri_escape_chars, pool);
 }
 
-const char uri_autoescape_chars[256] = {
+static const char uri_autoescape_chars[256] = {
   1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,
   1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,
   0, 1, 0, 1, 1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,
@@ -976,7 +1093,7 @@ svn_path_uri_decode(const char *path, apr_pool_t *pool)
   retstr = svn_stringbuf_create("", pool);
 
   /* avoid repeated realloc */
-  svn_stringbuf_ensure(retstr, strlen(path) + 1); 
+  svn_stringbuf_ensure(retstr, strlen(path) + 1);
 
   retstr->len = 0;
   for (i = 0; path[i]; i++)
@@ -1034,26 +1151,23 @@ svn_path_get_absolute(const char **pabsolute,
   apr_status_t apr_err;
   const char *path_apr;
 
-  SVN_ERR(svn_path_cstring_from_utf8
-          (&path_apr, svn_path_canonicalize(relative, pool), pool));
-
-  if (svn_path_is_url(path_apr))
+  if (svn_path_is_url(relative))
     {
-      buffer = apr_pstrdup(pool, path_apr);
+      *pabsolute = apr_pstrdup(pool, relative);
+      return SVN_NO_ERROR;
     }
-  else
-    {
-      apr_err = apr_filepath_merge(&buffer, NULL,
-                                   path_apr,
-                                   (APR_FILEPATH_NOTRELATIVE
-                                    | APR_FILEPATH_TRUENAME),
-                                   pool);
 
-      if (apr_err)
-        return svn_error_createf(SVN_ERR_BAD_FILENAME, NULL,
-                                 _("Couldn't determine absolute path of '%s'"), 
-                                 svn_path_local_style(relative, pool));
-    }
+  SVN_ERR(svn_path_cstring_from_utf8(&path_apr, relative, pool));
+
+  apr_err = apr_filepath_merge(&buffer, NULL,
+                               path_apr,
+                               APR_FILEPATH_NOTRELATIVE
+                               | APR_FILEPATH_TRUENAME,
+                               pool);
+  if (apr_err)
+    return svn_error_createf(SVN_ERR_BAD_FILENAME, NULL,
+                             _("Couldn't determine absolute path of '%s'"),
+                             svn_path_local_style(relative, pool));
 
   SVN_ERR(svn_path_cstring_to_utf8(pabsolute, buffer, pool));
   *pabsolute = svn_path_canonicalize(*pabsolute, pool);
@@ -1078,8 +1192,7 @@ svn_path_split_if_file(const char *path,
 
   if (err || finfo.filetype == APR_REG)
     {
-      if (err)
-        svn_error_clear(err);
+      svn_error_clear(err);
       svn_path_split(path, pdirectory, pfile, pool);
     }
   else if (finfo.filetype == APR_DIR)
@@ -1087,7 +1200,7 @@ svn_path_split_if_file(const char *path,
       *pdirectory = path;
       *pfile = SVN_EMPTY_PATH;
     }
-  else 
+  else
     {
       return svn_error_createf(SVN_ERR_BAD_FILENAME, NULL,
                                _("'%s' is neither a file nor a directory name"),
@@ -1136,7 +1249,7 @@ svn_path_canonicalize(const char *path, apr_pool_t *pool)
       if (!uri && *src == '/')
         *(dst++) = *(src++);
 #endif /* WIN32 or Cygwin */
-      
+
     }
 
   while (*src)
@@ -1171,7 +1284,7 @@ svn_path_canonicalize(const char *path, apr_pool_t *pool)
   /* Remove the trailing slash. */
   if ((canon_segments > 0 || uri) && *(dst - 1) == '/')
     dst--;
-  
+
   *dst = '\0';
 
 #if defined(WIN32) || defined(__CYGWIN__)
@@ -1185,6 +1298,13 @@ svn_path_canonicalize(const char *path, apr_pool_t *pool)
 }
 
 
+svn_boolean_t
+svn_path_is_canonical(const char *path, apr_pool_t *pool)
+{
+  return (strcmp(path, svn_path_canonicalize(path, pool)) == 0);
+}
+
+
 
 /** Get APR's internal path encoding. */
 static svn_error_t *
@@ -1239,6 +1359,65 @@ svn_path_cstring_to_utf8(const char **path_utf8,
     return svn_utf_cstring_to_utf8(path_utf8, path_apr, pool);
 }
 
+
+/* Return a copy of PATH, allocated from POOL, for which control
+   characters have been escaped using the form \NNN (where NNN is the
+   octal representation of the byte's ordinal value).  */
+static const char *
+illegal_path_escape(const char *path, apr_pool_t *pool)
+{
+  svn_stringbuf_t *retstr;
+  apr_size_t i, copied = 0;
+  int c;
+
+  retstr = svn_stringbuf_create("", pool);
+  for (i = 0; path[i]; i++)
+    {
+      c = (unsigned char)path[i];
+      if (! svn_ctype_iscntrl(c))
+        continue;
+
+      /* If we got here, we're looking at a character that isn't
+         supported by the (or at least, our) URI encoding scheme.  We
+         need to escape this character.  */
+
+      /* First things first, copy all the good stuff that we haven't
+         yet copied into our output buffer. */
+      if (i - copied)
+        svn_stringbuf_appendbytes(retstr, path + copied,
+                                  i - copied);
+
+      /* Now, sprintf() in our escaped character, making sure our
+         buffer is big enough to hold the '%' and two digits.  We cast
+         the C to unsigned char here because the 'X' format character
+         will be tempted to treat it as an unsigned int...which causes
+         problem when messing with 0x80-0xFF chars.  We also need space
+         for a null as sprintf will write one. */
+      /*### The backslash separator doesn't work too great with Windows,
+         but it's what we'll use for consistency with invalid utf8
+         formatting (until someone has a better idea) */
+      svn_stringbuf_ensure(retstr, retstr->len + 4);
+      sprintf(retstr->data + retstr->len, "\\%03o", (unsigned char)c);
+      retstr->len += 4;
+
+      /* Finally, update our copy counter. */
+      copied = i + 1;
+    }
+
+  /* If we didn't encode anything, we don't need to duplicate the string. */
+  if (retstr->len == 0)
+    return path;
+
+  /* Anything left to copy? */
+  if (i - copied)
+    svn_stringbuf_appendbytes(retstr, path + copied, i - copied);
+
+  /* retstr is null-terminated either by sprintf or the svn_stringbuf
+     functions. */
+
+  return retstr->data;
+}
+
 svn_error_t *
 svn_path_check_valid(const char *path, apr_pool_t *pool)
 {
@@ -1252,9 +1431,52 @@ svn_path_check_valid(const char *path, apr_pool_t *pool)
             (SVN_ERR_FS_PATH_SYNTAX, NULL,
              _("Invalid control character '0x%02x' in path '%s'"),
              *c,
-             svn_path_local_style(path, pool));
+             illegal_path_escape(svn_path_local_style(path, pool), pool));
         }
     }
 
   return SVN_NO_ERROR;
 }
+
+void
+svn_path_splitext(const char **path_root,
+                  const char **path_ext,
+                  const char *path,
+                  apr_pool_t *pool)
+{
+  const char *last_dot, *last_slash;
+
+  /* Easy out -- why do all the work when there's no way to report it? */
+  if (! (path_root || path_ext))
+    return;
+
+  /* Do we even have a period in this thing?  And if so, is there
+     anything after it?  We look for the "rightmost" period in the
+     string. */
+  last_dot = strrchr(path, '.');
+  if (last_dot && (last_dot + 1 != '\0'))
+    {
+      /* If we have a period, we need to make sure it occurs in the
+         final path component -- that there's no path separator
+         between the last period and the end of the PATH -- otherwise,
+         it doesn't count.  Also, we want to make sure that our period
+         isn't the first character of the last component. */
+      last_slash = strrchr(path, '/');
+      if ((last_slash && (last_dot > (last_slash + 1)))
+          || ((! last_slash) && (last_dot > path)))
+        {
+          if (path_root)
+            *path_root = apr_pstrmemdup(pool, path,
+                                        (last_dot - path + 1) * sizeof(*path));
+          if (path_ext)
+            *path_ext = apr_pstrdup(pool, last_dot + 1);
+          return;
+        }
+    }
+  /* If we get here, we never found a suitable separator character, so
+     there's no split. */
+  if (path_root)
+    *path_root = apr_pstrdup(pool, path);
+  if (path_ext)
+    *path_ext = "";
+}
